/*
** BeNES - Nintendo Entertaiment System Emulator for BeOS
**
** * 1999 by makoto yamagata
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef NESMAPPER4_H_DEFINED
#define NESMAPPER4_H_DEFINED

#include "NesAbstructMapper.h"

class NesMapper4: public NesAbstructMapper {
	ushort mChrOffset;
	bool mPrgHiBank;
	uchar mCommand;
	ushort mPrgBanks;
	ushort mChrBanks;
	ushort mScanline;
	int mIrqLatch;
	int mIrqCounter;
	bool mIrqEnabled;
	bool mIrqOccerEnabled;
	bool mIrqSetMode;
	uchar mBankRegs[8];
	enum { MAP4_PRGBANK_SIZE=0x2000, MAP4_CHRBANK_SIZE=0x400 };
protected:
	void bankChange(NesIOInterface* io, uchar value);
public:
	void init(NesIOInterface* io) {
		mChrOffset = 0x0000;
		mPrgHiBank = false;
		mCommand = 0;
		mIrqLatch = 0;
		mIrqCounter = 0;
		mIrqEnabled = false;
		mIrqSetMode = false;
		mIrqOccerEnabled = false;
		mScanline = 0;
		NesROMInterface* rom = io->getROM();
		mPrgBanks = rom->sizeofPRGROM() / MAP4_PRGBANK_SIZE;
		mChrBanks = rom->sizeofCHRROM() / MAP4_CHRBANK_SIZE;
		if (mChrBanks == 0) mChrBanks = mPrgBanks;
	}
	void enter(NesIOInterface* io, ushort address, uchar value);
	void scanLine(NesIOInterface* io) {
		mIrqSetMode = false;
		if (io->getPPU()->isDrawing()) {
			mIrqCounter--;
		}
		if (mIrqCounter <= 0) {
			if (mIrqEnabled && mIrqOccerEnabled) {
				io->getCPU()->intIRQ();
				mIrqOccerEnabled = false;
			}
			mIrqSetMode = true;
			mIrqCounter = mIrqLatch;
		}
	}
};

//------------------------------------------------------------
inline void NesMapper4::enter(
	NesIOInterface* io, ushort address, uchar value)
{
	NesPPUInterface* ppu = io->getPPU();
	uchar command_back;
	switch (address & 0xe001) {
	case 0x8000:
		mChrOffset = 0x0000;
		if (value & 0x80) {
			mChrOffset = 0x1000;
		}
		mPrgHiBank = ((value & 0x40) != 0);
		command_back = mCommand;
		if ((mCommand&0x80) != (value&0x80)) {
			for (int i=0; i<6; i++) {
				mCommand = i;
				bankChange(io, mBankRegs[i]);
			}
		}
		mCommand = command_back;
		if ((mCommand&0x40) != (value&0x40)) {
			for (int i=6; i<8; i++) {
				mCommand = i;
				bankChange(io, mBankRegs[i]);
			}
		}
		mCommand = value;
		break;
	case 0x8001:
		bankChange(io, value);
		break;
	case 0xa000:
		if (value & 0x40) {
			ppu->setMirrorType(PPU_MIRROR_ONEHIGH);
		}
		else if (value & 0x01) {
			ppu->setMirrorType(PPU_MIRROR_HORIZONTAL);
		}
		else {
			ppu->setMirrorType(PPU_MIRROR_VERTICAL);
		}
		break;
	case 0xc000:
		mIrqLatch = value;
		if (mIrqSetMode) {
			mIrqCounter = mIrqLatch;
		}
		break;
	case 0xc001:
		mIrqSetMode = true;
		mIrqCounter = value;
		break;
	case 0xe000:
		mIrqEnabled = false;
		break;
	case 0xe001:
		mIrqEnabled = true;
		mIrqOccerEnabled = true;
		break;
	}
}

//------------------------------------------------------------
inline void NesMapper4::bankChange(NesIOInterface* io, uchar value)
{
	NesCPUInterface* cpu = io->getCPU();
	NesROMInterface* rom = io->getROM();
	NesPPUInterface* ppu = io->getPPU();
	mBankRegs[mCommand&0x07] = value;
	int vram_offset = 0x0000;
	int vrom_offset = 0x0000;
	int chr_size = 0;
	int pram_offset = 0x0000;
	int prom_offset = 0x0000;
	int prg_size = 0;
	switch (mCommand&0x07) {
	case 0:
		vram_offset = 0x0000 ^ mChrOffset;
		value &= 0xfe;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE * 2;
		break;
	case 1:
		vram_offset = 0x0800 ^ mChrOffset;
		value &= 0xfe;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE * 2;
		break;
	case 2:
		vram_offset = 0x1000 ^ mChrOffset;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE;
		break;
	case 3:
		vram_offset = 0x1400 ^ mChrOffset;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE;
		break;
	case 4:
		vram_offset = 0x1800 ^ mChrOffset;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE;
		break;
	case 5:
		vram_offset = 0x1c00 ^ mChrOffset;
		value %= (mChrBanks);
		vrom_offset = value * MAP4_CHRBANK_SIZE;
		chr_size = MAP4_CHRBANK_SIZE;
		break;
	case 6:
		prg_size = MAP4_PRGBANK_SIZE;
		value %= (mPrgBanks);
		prom_offset = value * MAP4_PRGBANK_SIZE;
		if (mPrgHiBank) pram_offset = 0xc000;
		else pram_offset = 0x8000;
		break;
	case 7:
		prg_size = MAP4_PRGBANK_SIZE;
		value %= (mPrgBanks);
		prom_offset = value * MAP4_PRGBANK_SIZE;
		pram_offset = 0xa000;
		break;
	}
	if (chr_size) {
		ppu->memmove(vram_offset, 
			rom->getCHRROM(vrom_offset), chr_size);
	}
	if (prg_size) {
		cpu->memmove(pram_offset, 
			rom->getPRGROM(prom_offset), prg_size);
		uchar* mem = rom->getPRGROM(rom->sizeofPRGROM()-(MAP4_PRGBANK_SIZE*2));
		if (mPrgHiBank) {
			cpu->memmove(0x8000, mem, MAP4_PRGBANK_SIZE);
		} else {
			cpu->memmove(0xc000, mem, MAP4_PRGBANK_SIZE);
		}
	}
}


#endif
